@@ -59,15 +59,20 @@ sub new {
     $self->{func} = $new{Callbacks};
     $self->{global_timeout} = $new{GlobalTimeout}||120;
     $self->{last_active}=time;
+    $self->{timer_actions}=[];
     $self->{defaults}={
                        threads=>$new{Threads}||2,
                        port=>$new{Port}||25,
                        timeout=>$new{Timeout}||30,
                        secure=>$new{Secure}||0,
+		       sleep=>{
+                               hang=>0,
+                               fail=>0,
+                               },
                        retry=>{
                         global_hang=>1,
                         hang=>1,
-                        fail=>0
+                        fail=>5
                        }
                        };
     
@@ -219,6 +224,7 @@ sub quit {
                                         
                                         if ($end == 1) {
                                             undef($timer);
+                                            undef($timer2);
                                             $self->{cv}->send;
                                         }
                                         
@@ -251,17 +257,37 @@ sub quit {
     }
     
 
-    
+    my $timer2   = AnyEvent->timer(
+    	interval => 1,
+    	cb       => sub {
+       	my @actions;
+		foreach my $i ( 0 .. $#{ $self->{timer_actions} } ) {
+			$self->{timer_actions}[$i][0]++;
+                        #print "OK(  $self->{timer_actions}[$i][0] >= $self->{timer_actions}[$i][1] )\n";
+			if ( $self->{timer_actions}[$i][0] >= $self->{timer_actions}[$i][1] ) {
+				&{ $self->{timer_actions}[$i][2] }( @{ $self->{timer_actions}[$i][3] } );
+			
+			} else {
+				push(@actions,$self->{timer_actions}[$i]);
+			}
+		
+		}
+		$self->{timer_actions}=\@actions;
+
+	},
+    );    
     
     #$self->_BULK();
     $self->{cv}->recv if $total > 0;
      
     undef($timer) if defined($timer);
+    undef($timer2) if defined($timer2);
      
-     
-     
+    my $rt1=$self->_FUNC('complete',$self,[-1,-1],0,[]);
     foreach my $h ( keys(%{ $self->{threads} })  ) {
         foreach my $t ( 0..$self->{threads}{ $h } ) {
+            
+            my $rt2=$self->_FUNC('end_thread',$self,$k,0,[]);
             $self->{fh}{threads}{ $h }{ $t }->destroy if defined($self->{fh}{threads}{ $h }{ $t });
             
              $self->_DEBUG([$h,$t],'End Queue : (PASS:'.($self->{stats}{ $h }{ $t }{queue}{pass}).'|HANG:'.($self->{stats}{ $h }{ $t }{queue}{hang}).'|FAIL:'.($self->{stats}{ $h }{ $t }{queue}{fail}).'|TOTAL:'.$self->{stats}{ $h }{ $t }{queue}{total}.')',5) if $self->{debug} >= 1;
@@ -319,6 +345,9 @@ sub _PREPARE {
         $self->{helo}{ $new{Host} }=$new{Hello}||$self->{host}{ $new{Host} };
         $self->{timeout}{ $new{Host} }=$new{Timeout}||$self->{defaults}{timeout};
         $self->{threads}{ $new{Host} }=($new{Threads}||$self->{defaults}{threads}) - 1;
+
+        $self->{sleep}{ $new{Host} }{hang}=$new{Sleep}{Hang}||$self->{defaults}{sleep}{hang};
+        $self->{sleep}{ $new{Host} }{fail}=$new{Sleep}{Fail}||$self->{defaults}{sleep}{fail};
         
         $self->{retry}{ $new{Host} }{hang}=$new{Retry}{Hang}||$self->{defaults}{retry}{hang};
         $self->{retry}{ $new{Host} }{global_hang}=$new{Retry}{GlobalHang}||$self->{defaults}{retry}{global_hang};
@@ -401,15 +430,30 @@ sub _CONNECT {
       on_timeout=> sub {
         $self->{stats}{ $k->[0] }{ $k->[1] }{queue}{hang}++;
         my $r=$self->_FUNC('hang',$self,$k,0,[$self->{on_queue}{ $k->[0] }{ $k->[1] }]);
-        $self->_DEBUG($k,'Email : '.(++$self->{stats}{ $k->[0] }{ $k->[1] }{queue}{count}).' : '.$self->{on_queue}{ $k->[0] }{ $k->[1] }{to}.' : HANG',8) if $self->{debug} >= 1;
-        $self->reconnect($k, $self->{retry}{ $k->[0] }{hang} ) if ($r == 1 or $r == 103);
+        $self->_DEBUG($k,'Email : '.(++$self->{stats}{ $k->[0] }{ $k->[1] }{queue}{count}).' : '.$self->{on_queue}{ $k->[0] }{ $k->[1] }{to}.' : HANG : '.$self->{stage}{ $k->[0] }{ $k->[1] }[0].','.$self->{stage}{ $k->[0] }{ $k->[1] }[1], 7) if $self->{debug} >= 1;
+
+	 if ($r == 1 or $r == 103 ) {
+	 	if ( ($self->{sleep}{ $k->[0] }{hang}||0) > 0 ) {
+	       	push(@{$self->{timer_actions}},[0,$self->{sleep}{ $k->[0] }{hang},'reconnect',[$self, $k, $self->{retry}{ $k->[0] }{hang}]] );
+	 	} else {
+	       	$self->reconnect($k, $self->{retry}{ $k->[0] }{hang} );
+		}
+
+	 }
+        
       },
       on_error => sub {
         $self->{stats}{ $k->[0] }{ $k->[1] }{queue}{fail}++;
         my $r=$self->_FUNC('fail',$self,$k,0,[$self->{on_queue}{ $k->[0] }{ $k->[1] }]);
         
-        $self->_DEBUG($k,'Email : '.(++$self->{stats}{ $k->[0] }{ $k->[1] }{queue}{count}).' : '.$self->{on_queue}{ $k->[0] }{ $k->[1] }{to}.' : FAIL',8) if $self->{debug} >= 1;
-        $self->reconnect($k, $self->{retry}{ $k->[0] }{fail} ) if ($r == 1 or $r == 103);
+        $self->_DEBUG($k,'Email : '.(++$self->{stats}{ $k->[0] }{ $k->[1] }{queue}{count}).' : '.$self->{on_queue}{ $k->[0] }{ $k->[1] }{to}.' : FAIL : '.$self->{stage}{ $k->[0] }{ $k->[1] }[0].','.$self->{stage}{ $k->[0] }{ $k->[1] }[1], 7) if $self->{debug} >= 1;
+
+	if ( ($self->{sleep}{ $k->[0] }{fail}||0) > 0 ) {
+	    push(@{$self->{timer_actions}},[0,$self->{sleep}{ $k->[0] }{fail},'reconnect',[$self, $k, $self->{retry}{ $k->[0] }{fail}]] );
+	} else {
+	    $self->reconnect($k, $self->{retry}{ $k->[0] }{fail} );
+	}
+
         
         },
       %extra
@@ -685,7 +729,7 @@ sub _READ {
                    
                     if ( ($self->{secure}{ $k->[0] } == 3 or $self->{secure}{ $k->[0] } == 4) and exists($self->{header}{ $k->[0] }{ $k->[1] }{starttls})  ) {
                        $self->_STARTTLS($k);
-                    } elsif ( exists($self->{auth}{ $k->[0] }{ $k->[1] }) ) {
+                    } elsif ( ($self->{auth}{ $k->[0] }{ $k->[1] }[0]||'0') ne '0' ) {
                         $self->_AUTH($k);
  
                     } else {
